@@ -21,7 +21,9 @@ __title__ = 'pronotepy'
 __author__ = 'bain, Xiloe'
 __license__ = 'MIT'
 __copyright__ = 'Copyright (c) bain, Xiloe'
-__version__ = '1.2.2'
+__version__ = '1.4.0'
 
-from .pronoteAPI import Client, ParentClient
-from .dataClasses import *
+from .dataClasses import Util, Subject, Period, File, Grade, Lesson, Message, LessonContent, ClientInfo, \
+    Average, Acquisition, Evaluation, Homework, StudentClass, Student, Identity, Guardian
+from .clients import Client, ParentClient, VieScolaireClient
+from .exceptions import *
