@@ -1,13 +1,14 @@
-import logging
 import datetime
-import re
 import json
-from urllib.parse import quote
-from Crypto.Util import Padding
+import logging
+import re
 from html import unescape
 from typing import Union, List
+from urllib.parse import quote
 
+from Crypto.Util import Padding
 
+from .exceptions import IncorrectJson
 
 log = logging.getLogger(__name__)
 log.setLevel(logging.DEBUG)
@@ -52,7 +53,14 @@ class Util:
                 for level in actual_dict:
                     out = out[level]
             except KeyError:
-                output[attribute_dict[key][0]] = None
+                if len(attribute_dict[key]) == 3:
+                    output[attribute_dict[key][0]] = attribute_dict[key][2]
+                else:
+                    try:
+                        output[attribute_dict[key][0]] = attribute_dict[key][1](None)
+                    except Exception as e:
+                        log.debug("Exception while parsing json (" + str(actual_dict) + "), setting to None: " + str(e))
+                        output[attribute_dict[key][0]] = None
             else:
                 output[attribute_dict[key][0]] = attribute_dict[key][1](out)
         return output
@@ -64,6 +72,10 @@ class Util:
         else:
             return string
 
+    @staticmethod
+    def date_parse(formatted_date: str):
+        return datetime.datetime.strptime(formatted_date, '%d/%m/%Y').date()
+
 
 class Subject:
     """
@@ -102,9 +114,9 @@ class Period:
         the id of the period (used internally)
     name : str
         name of the period
-    start : str
+    start : datetime.datetime
         date on which the period starts
-    end : str
+    end : datetime.datetime
         date on which the period ends
     """
 
@@ -127,7 +139,7 @@ class Period:
     @property
     def grades(self):
         """Get grades from the period."""
-        json_data = {'Periode': {'N': self.id, 'L': self.name}}
+        json_data = {'periode': {'N': self.id, 'L': self.name}}
         response = self._client.post('DernieresNotes', 198, json_data)
         log.debug(response)
         grades = response['donneesSec']['donnees']['listeDevoirs']['V']
@@ -137,7 +149,7 @@ class Period:
     def averages(self):
         """Get averages from the period."""
 
-        json_data = {'Periode': {'N': self.id, 'L': self.name}}
+        json_data = {'periode': {'N': self.id, 'L': self.name}}
         response = self._client.post('DernieresNotes', 198, json_data)
         crs = response['donneesSec']['donnees']['listeServices']['V']
         return [Average(c) for c in crs]
@@ -146,7 +158,7 @@ class Period:
     def overall_average(self):
         """Get overall average from the period. If the period average is not provided by pronote, then it's calculated.
         Calculation may not be the same as the actual average. (max difference 0.01)"""
-        json_data = {'Periode': {'N': self.id, 'L': self.name}}
+        json_data = {'periode': {'N': self.id, 'L': self.name}}
         response = self._client.post('DernieresNotes', 198, json_data)
         average = response['donneesSec']['donnees'].get('moyGenerale')
         if average:
@@ -174,7 +186,7 @@ class Period:
 
     @property
     def evaluations(self):
-        json_data = {'Periode': {'N': self.id, 'L': self.name, 'G': 2}}
+        json_data = {'periode': {'N': self.id, 'L': self.name, 'G': 2}}
         response = self._client.post('DernieresEvaluations', 201, json_data)
         evaluations = response['donneesSec']['donnees']['listeEvaluations']['V']
         return [Evaluation(e) for e in evaluations]
@@ -213,7 +225,7 @@ class Average:
     attribute_guide = {
         'moyEleve,V': ('student', Util.grade_parse),
         'baremeMoyEleve,V': ('out_of', Util.grade_parse),
-        'baremeMoyEleveParDefault,V': ('default_out_of', Util.grade_parse),
+        'baremeMoyEleveParDefault,V': ('default_out_of', Util.grade_parse, ""),
         'moyClasse,V': ('class_average', Util.grade_parse),
         'moyMin,V': ('min', Util.grade_parse),
         'moyMax,V': ('max', Util.grade_parse)
@@ -276,8 +288,8 @@ class Grade:
         "N": ("id", str),
         "note,V": ("grade", Util.grade_parse),
         "bareme,V": ("out_of", Util.grade_parse),
-        "baremeParDefault,V": ("default_out_of", Util.grade_parse),
-        "date,V": ("date", lambda d: datetime.datetime.strptime(d, '%d/%m/%Y').date()),
+        "baremeParDefault,V": ("default_out_of", Util.grade_parse, ""),
+        "date,V": ("date", Util.date_parse),
         "service,V": ("subject", Subject),
         "periode,V,N": ("period", lambda p: Util.get(Period.instances, id=p)),
         "moyenne,V": ("average", Util.grade_parse),
@@ -325,12 +337,16 @@ class Lesson:
         background color of the lesson
     outing : bool
         if it is a pedagogical outing
-    start : str
+    start : datetime.datetime
         starting time of the lesson
+    end : datetime.datetime
+        end of the lesson
     group_name : str
         Name of the group.
     exempted : bool
         Specifies if the student's presence is exempt.
+    virtual_classrooms : List[str]
+        List of urls for virtual classrooms
     """
 
     id: str
@@ -344,20 +360,23 @@ class Lesson:
     start: datetime.datetime
     group_name: str
     exempted: bool
+    virtual_classrooms: List[str]
+    end: datetime.datetime
 
     __slots__ = ['id', 'subject', 'teacher_name', 'classroom', 'start',
                  'canceled', 'status', 'background_color', 'detention',
                  'end', 'outing', 'group_name', 'student_class', 'exempted',
-                 '_client', '_content']
+                 '_client', '_content', 'virtual_classrooms']
     attribute_guide = {
         'DateDuCours,V': ('start', lambda d: datetime.datetime.strptime(d, '%d/%m/%Y %H:%M:%S')),
         'N': ('id', str),
         'estAnnule': ('canceled', bool),
         'Statut': ('status', str),
         'CouleurFond': ('background_color', str),
-        'estRetenue': ('detention', bool),
-        'estSortiePedagogique': ('outing', bool),
-        'dispenseEleve': ('exempted', bool)
+        'estRetenue': ('detention', bool, False),
+        'estSortiePedagogique': ('outing', bool, False),
+        'dispenseEleve': ('exempted', bool, False),
+        'listeVisios,V': ('virtual_classrooms', lambda l: [i["url"] for i in l], [])
     }
     transformers = {
         16: ('subject', Subject),
@@ -377,14 +396,17 @@ class Lesson:
         # Pronote gives us the place where the hour should be in a week, when we modulo that with the amount of
         # hours in a day we can get the "place" when the hour starts. Then we just add the duration (and substract 1)
         end_place = parsed_json['place'] % (len(
-            client.func_options['donneesSec']['donnees']['General']['ListeHeuresFin']['V'])-1) + parsed_json['duree']-1
+            client.func_options['donneesSec']['donnees']['General']['ListeHeuresFin']['V']) - 1) + parsed_json[
+                        'duree'] - 1
 
         # With the end "place" now known we can look up the ending time in func_options
         self.end = None
-        for end_time in client.func_options['donneesSec']['donnees']['General']['ListeHeuresFin']['V']:
-            if end_time['G'] == end_place:
-                end_time = datetime.datetime.strptime(end_time['L'], "%Hh%M").time()
-                self.end = self.start.replace(hour=end_time.hour, minute=end_time.minute)
+        end_time = next(filter(
+            lambda x: x['G'] == end_place,
+            client.func_options['donneesSec']['donnees']['General']['ListeHeuresFin']['V']
+        ))
+        end_time = datetime.datetime.strptime(end_time['L'], "%Hh%M").time()
+        self.end = self.start.replace(hour=end_time.hour, minute=end_time.minute)
 
         self.teacher_name = self.classroom = self.group_name = self.student_class = ''
         self.subject = None
@@ -491,7 +513,7 @@ class File:
         for key in prepared_json:
             self.__setattr__(key, prepared_json[key])
         self._client = client
-        padd = Padding.pad(json.dumps({'N': self.id, 'Actif': True}).replace(' ', '').encode(), 16)
+        padd = Padding.pad(json.dumps({'N': self.id, 'G': 1}).replace(' ', '').encode(), 16)
         magic_stuff = client.communication.encryption.aes_encrypt(padd).hex()
         self.url = client.communication.root_site \
                    + '/FichiersExternes/' \
@@ -543,7 +565,7 @@ class Homework:
         the background color of the homework 
     done : bool
         if the homework is marked done
-    date : str
+    date : datetime.date
         deadline
     """
 
@@ -561,7 +583,7 @@ class Homework:
         'TAFFait': ('done', bool),
         'Matiere,V': ('subject', Subject),
         'CouleurFond': ('background_color', str),
-        'PourLe,V': ('date', lambda d: datetime.datetime.strptime(d, '%d/%m/%Y').date()),
+        'PourLe,V': ('date', Util.date_parse),
         'ListePieceJointe,V': ('_files', tuple)
     }
 
@@ -607,7 +629,7 @@ class Message:
         Recipitents of the message. ! May be just ['# recipients'] !
     seen : bool
         if the message was seen
-    date : str
+    date : datetime.datetime
         the date when the message was sent
     """
 
@@ -615,7 +637,7 @@ class Message:
     author: str
     recipients: list
     seen: bool
-    date: str
+    date: datetime.datetime
 
     __slots__ = ['id', 'author', 'recipients', 'seen', 'date', '_client', '_listePM']
     attribute_guide = {
@@ -806,7 +828,7 @@ class Evaluation:
         # Can we just appreciate the readability of this?
         'listeNiveauxDAcquisitions,V': (
             'acquisitions', lambda x: sorted([Acquisition(y) for y in x], key=lambda z: z.order)),
-        "date,V": ("date", lambda d: datetime.datetime.strptime(d, '%d/%m/%Y').date()),
+        "date,V": ("date", Util.date_parse),
     }
 
     def __init__(self, json_):
@@ -815,13 +837,241 @@ class Evaluation:
             self.__setattr__(key, prepared_json[key])
 
 
-class DataError(Exception):
+class Identity:
     """
-    Base exception for any errors made by creating or manipulating data classes.
+    Represents an Identity of a person
+    
+    Attributes
+    ----------
+    postal_code: str
+    date_of_birth: datetime.date
+    email: str
+    last_name: str
+    country: str
+    mobile_number: str
+    landline_number: str
+    other_phone_number: str
+    city: str
+    place_of_birth: str
+    first_names: List[str]
+    address: List[str]
+    formatted_address: str
+        concatenated address information into a single string
     """
-    pass
+    attribute_guide = {
+        "CP": ("postal_code", str),
+        "dateNaiss": ("date_of_birth", Util.date_parse, None),
+        "email": ("email", str),
+        "nom": ("last_name", str),
+        "pays": ("country", str),
+        "telPort": ("mobile_number", str),
+        "telFixe": ("landline_number", str),
+        "telAutre": ("other_phone_number", str),
+        "ville": ("city", str),
+        "villeNaiss": ("place_of_birth", str)
+    }
+    __slots__ = ["postal_code", "date_of_birth", "email", "last_name", "country", "mobile_number", "landline_number",
+                 "other_phone_number", "city", "place_of_birth", "first_names", "address", "formatted_address"]
+
+    postal_code: str
+    date_of_birth: datetime.date
+    email: str
+    last_name: str
+    country: str
+    mobile_number: str
+    landline_number: str
+    other_phone_number: str
+    city: str
+    place_of_birth: str
+    first_names: List[str]
+    address: List[str]
+    formatted_address: str
 
 
-class IncorrectJson(DataError):
-    """Bad json"""
-    pass
+    def __init__(self, json_):
+        prepared_json = Util.prepare_json(self.__class__, json_)
+        for key in prepared_json:
+            self.__setattr__(key, prepared_json[key])
+
+        self.address = []
+        i = 1
+        while True:
+            option = json_.get("adresse" + str(i))
+            if not option: break
+            self.address.append(option)
+            i += 1
+        self.formatted_address = ','.join([*self.address, self.postal_code, self.city, self.country])
+        self.first_names = [json_.get("prenom"), json_.get("prenom2"), json_.get("prenom3")]
+
+
+class Guardian:
+    """
+    Represents a guardian of a student.
+
+    Attributes
+    ----------
+    identity: dataClasses.Identity
+    accepteInfosProf: bool
+    authorized_email: bool
+    authorized_pick_up_kid: bool
+    urgency_contact: bool
+    preferred_responsible_contact: bool
+    accomodates_kid: bool
+    relatives_link: str
+    responsibility_level: str
+    financially_responsible: bool
+    full_name: str
+    is_legal: bool
+    """
+    attribute_guide = {
+        "accepteInfosProf": ("accepteInfosProf", bool),
+        "autoriseEmail": ("authorized_email", bool),
+        "autoriseRecupererEnfant": ("authorized_pick_up_kid", bool),
+        "contactUrgence": ("urgency_contact", bool),
+        "estResponsablePreferentiel": ("preferred_responsible_contact", bool),
+        "hebergeEnfant": ("accomodates_kid", bool),
+        "lienParente": ("relatives_link", str),
+        "niveauResponsabilite": ("responsibility_level", str),
+        "responsableFinancier": ("financially_responsible", bool),
+        "nom": ("full_name", str)
+    }
+    __slots__ = ["identity", "accepteInfosProf", "authorized_email", "authorized_pick_up_kid", "urgency_contact",
+                 "preferred_responsible_contact", "accomodates_kid", "relatives_link", "responsibility_level",
+                 "financially_responsible", "full_name", "is_legal"]
+
+    identity: Identity
+    accepteInfosProf: bool
+    authorized_email: bool
+    authorized_pick_up_kid: bool
+    urgency_contact: bool
+    preferred_responsible_contact: bool
+    accomodates_kid: bool
+    relatives_link: str
+    responsibility_level: str
+    financially_responsible: bool
+    full_name: str
+    is_legal: bool
+
+    def __init__(self, json_):
+        prepared_json = Util.prepare_json(self.__class__, json_)
+        for key in prepared_json:
+            self.__setattr__(key, prepared_json[key])
+
+        self.identity = Identity(json_)
+        self.is_legal = self.responsibility_level == "LEGAL"
+
+
+class Student:
+    """
+    Represents a student
+
+    Attributes
+    ----------
+    full_name: str
+    id: str
+    enrollment_date: datetime.date
+    date_of_birth: datetime.date
+    projects: List[str]
+    last_name: str
+    first_names: str
+    sex: str
+    options: List[str]
+        language options
+    """
+    attribute_guide = {
+        "L": ("full_name", str),
+        "N": ("id", str),
+        "entree,V": ("enrollment_date", Util.date_parse),
+        "neLe,V": ("date_of_birth", Util.date_parse),
+        "listeProjets,V": ("projects", lambda p: [f"{x['typeAmenagement']} ({x['handicap']})" for x in p]),
+        "nom": ("last_name", str),
+        "prenoms": ("first_names", str),
+        "sexe": ("sex", str),
+    }
+    __slots__ = ["_client", "full_name", "id", "enrollment_date", "date_of_birth", "projects", "last_name",
+                 "first_names", "sex", "options", "_cache"]
+
+    full_name: str
+    id: str
+    enrollment_date: datetime.date
+    date_of_birth: datetime.date
+    projects: List[str]
+    last_name: str
+    first_names: str
+    sex: str
+    options: List[str]
+
+    def __init__(self, client, json_):
+        prepared_json = Util.prepare_json(self.__class__, json_)
+        for key in prepared_json:
+            self.__setattr__(key, prepared_json[key])
+
+        self._client = client
+        self._cache = None
+
+        self.options = []
+        i = 1
+        while True:
+            option = json_.get("option" + str(i))
+            if not option:
+                break
+            self.options.append(option)
+            i += 1
+
+    @property
+    def identity(self) -> Identity:
+        """
+        Identity of this student
+        """
+        if not self._cache:
+            self._cache = self._client.post("FicheEleve", 105,
+                                            {"Eleve": {"N": self.id}, "AvecEleve": True, "AvecResponsables": True})
+        return Identity(self._cache["donneesSec"]["donnees"]["Identite"])
+
+    @property
+    def guardians(self) -> List[Guardian]:
+        """
+        List of responsible persons (parents).
+        """
+        if not self._cache:
+            self._cache = self._client.post("FicheEleve", 105, {"N": self.id})
+        return [Guardian(j) for j in self._cache["donneesSec"]["donnees"]["Responsables"]["V"]]
+
+
+class StudentClass:
+    """
+    Represents a class of students
+
+    Attributes
+    ----------
+    name: str
+    id: str
+    responsible: bool
+        is the teacher responsible for the class
+    grade: str
+    """
+    attribute_guide = {
+        "L": ("name", str),
+        "N": ("id", str),
+        "estResponsable": ("responsible", bool),
+        "niveau,V,L": ("grade", str, "")
+    }
+    __slots__ = ["name", "id", "responsible", "grade", "_client"]
+
+    name: str
+    id: str
+    responsible: bool
+    grade: str
+
+    def __init__(self, client, json_):
+        prepared_json = Util.prepare_json(self.__class__, json_)
+        for key in prepared_json:
+            self.__setattr__(key, prepared_json[key])
+
+        self._client = client
+
+    def students(self, period: Period = None) -> List[Student]:
+        period = period or self._client.periods[0]
+        r = self._client.post("ListeRessources", 105,
+                              {"classe": {"N": self.id, "G": 1}, "periode": {"N": period.id, "G": 1}})
+        return [Student(self._client, j) for j in r["donneesSec"]["donnees"]["listeRessources"]["V"]]
